"""Calcuate game rankings."""

import logging
from pathlib import Path
from typing import Union

import turicreate as tc

from board_game_recommender.recommend import BGGRecommender
from board_game_recommender.trust import user_trust

LOGGER = logging.getLogger(__name__)


def calculate_rankings(
    recommender: Union[BGGRecommender, Path, str],
    path_ratings: Union[Path, str],
    top: int = 100,
    min_ratings: int = 10,
) -> tc.SFrame:
    """Calcuate game rankings."""

    recommender = (
        BGGRecommender.load(recommender)
        if isinstance(recommender, (Path, str))
        else recommender
    )

    LOGGER.info("Using <%s> for recommendations", recommender)

    columns = BGGRecommender.columns_ratings
    columns["updated_at"] = str

    LOGGER.info("Loading ratings from <%s>", path_ratings)
    ratings = BGGRecommender.load_ratings_json(
        ratings_json=path_ratings,
        columns=columns,
    )
    LOGGER.info("Loaded %d ratings", len(ratings))

    users = ratings.groupby(
        key_column_names="bgg_user_name",
        operations={"ratings_count": tc.aggregate.COUNT()},
    )
    LOGGER.info("Found %d users in total", len(users))

    trust = user_trust(ratings=ratings, min_ratings=min_ratings)
    LOGGER.info("Calculated trust scores for %d users", len(trust))

    games = ratings.groupby(
        key_column_names="bgg_id",
        operations={
            "avg_rating": tc.aggregate.MEAN("bgg_user_rating"),
            "num_votes": tc.aggregate.COUNT(),
        },
    )
    LOGGER.info("Found %d games in total", len(games))
    del ratings

    if recommender.compilations is not None:
        games = games.filter_by(
            values=recommender.compilations,
            column_name="bgg_id",
            exclude=True,
        )
        LOGGER.info(
            "Restrict recommendations to %d games after removing compilations",
            len(games),
        )

    users = users.join(
        right=trust,
        on="bgg_user_name",
        how="inner",
    )
    del trust

    heavy_users = users[(users["ratings_count"] >= min_ratings) & (users["trust"] > 0)]
    LOGGER.info(
        "Got user info for %d users, %d of which are heavy users",
        len(users),
        len(heavy_users),
    )
    del users

    recommendations = recommender.model.recommend(
        users=heavy_users["bgg_user_name"],
        items=games["bgg_id"],
        exclude_known=False,
        k=top,
    )
    standard_recommendations = recommender.model.recommend(
        users=[None],
        items=games["bgg_id"],
        exclude_known=False,
        k=len(games),
    )["bgg_id", "score"]
    standard_recommendations.rename({"score": "score_standard"}, inplace=True)
    LOGGER.info(
        "Calculated a total of %d recommendations",
        len(recommendations) + len(standard_recommendations),
    )
    del recommender

    recommendations = recommendations.join(heavy_users, on="bgg_user_name", how="inner")
    recommendations["rev_rank"] = top + 1 - recommendations["rank"]
    recommendations["rev_rank_weighted"] = (
        recommendations["rev_rank"] * recommendations["trust"]
    )

    scores = recommendations.groupby(
        key_column_names="bgg_id",
        operations={
            "score": tc.aggregate.SUM("rev_rank"),
            "score_weighted": tc.aggregate.SUM(
                "rev_rank_weighted",
            ),
        },
    )
    del recommendations

    total_weight = heavy_users["trust"].sum()
    scores["score"] = scores["score"] / len(heavy_users)
    scores["score_weighted"] = scores["score_weighted"] / total_weight
    LOGGER.info("Calculated ranking scores for %d games", len(scores))
    del heavy_users

    scores = scores.join(
        right=standard_recommendations,
        on="bgg_id",
        how="outer",
    ).join(
        right=games,
        on="bgg_id",
        how="outer",
    )
    LOGGER.info("Calculated different scores for a total of %d games", len(scores))
    del games, standard_recommendations

    for col in ("score", "score_weighted", "score_standard", "avg_rating", "num_votes"):
        scores = scores.fillna(col, 0)

    # TODO what to do with ties?
    scores = scores.sort(
        key_column_names=[
            "score_weighted",
            "score",
            "score_standard",
            "avg_rating",
            "num_votes",
        ],
        ascending=False,
    )
    scores["rank_weighted"] = range(1, len(scores) + 1)

    scores = scores.sort(
        key_column_names=[
            "score",
            "score_weighted",
            "score_standard",
            "avg_rating",
            "num_votes",
        ],
        ascending=False,
    )
    scores["rank"] = range(1, len(scores) + 1)

    return scores



"""Baseline recommender models."""

import logging
import os
from collections import defaultdict
from typing import Any, Dict, FrozenSet, Iterable, List, Optional, Tuple, Union

import numpy as np
import pandas as pd

from board_game_recommender.base import BaseGamesRecommender

LOGGER = logging.getLogger(__name__)
PATH = Union[str, os.PathLike]


def dataframe_from_scores(
    columns: List[Any],
    index: Iterable[Any],
    scores: np.ndarray,
) -> pd.DataFrame:
    """Creates a Pandas DataFrame out of raw recommendation scores."""

    result = pd.DataFrame(
        index=list(index),
        columns=pd.MultiIndex.from_product([columns, ["score"]]),
        data=scores.T,
    )
    result[pd.MultiIndex.from_product([columns, ["rank"]])] = result.rank(
        method="min",
        ascending=False,
    ).astype(int)

    if len(columns) == 1:
        result.sort_values((columns[0], "rank"), inplace=True)

    return result[pd.MultiIndex.from_product([columns, ["score", "rank"]])]


class RandomGamesRecommender(BaseGamesRecommender):
    """Random recommender."""

    def __init__(self) -> None:
        self.rng = np.random.default_rng()

    @property
    def known_games(self) -> FrozenSet[int]:
        return frozenset()

    @property
    def rated_games(self) -> FrozenSet[int]:
        return frozenset()

    @property
    def known_users(self) -> FrozenSet[str]:
        return frozenset()

    def _recommendation_scores(self, users: int, games: int) -> np.ndarray:
        """Random scores."""
        return self.rng.random((users, games))

    def recommend(
        self,
        users: Iterable[str],
        games: Iterable[int],
        **kwargs,
    ) -> pd.DataFrame:
        """Random recommendations for certain users."""

        users = list(users)
        games = list(games)
        scores = self._recommendation_scores(users=len(users), games=len(games))

        return dataframe_from_scores(users, games, scores)

    def recommend_as_numpy(
        self,
        users: Iterable[str],
        games: Iterable[int],
    ) -> np.ndarray:
        """Random recommendations for certain users and games as a numpy array."""
        users = list(users)
        games = list(games)
        return self._recommendation_scores(users=len(users), games=len(games))

    def recommend_group(
        self,
        users: Iterable[str],
        games: Iterable[int],
        **kwargs,
    ) -> pd.DataFrame:
        """Random recommendations for a group of users."""

        games = list(games)
        scores = self._recommendation_scores(users=1, games=len(games))

        return dataframe_from_scores(["_all"], games, scores)

    def recommend_group_as_numpy(
        self,
        users: Iterable[str],
        games: Iterable[int],
    ) -> np.ndarray:
        """Random recommendations for a group of users and games as a numpy array."""
        games = list(games)
        return self._recommendation_scores(users=1, games=len(games))

    def recommend_similar(self, games: Iterable[int], **kwargs) -> pd.DataFrame:
        raise NotImplementedError

    def similar_games(self, games: Iterable[int], **kwargs) -> pd.DataFrame:
        raise NotImplementedError


class PopularGamesRecommender(BaseGamesRecommender):
    """Popular games recommender."""

    id_field: str = "bgg_id"
    user_id_field: str = "bgg_user_name"
    rating_id_field: str = "bgg_user_rating"

    scores: Dict[int, float]
    raw_scores: np.ndarray
    default_value: float
    game_ids: Tuple[int, ...]

    _known_games: Optional[FrozenSet[int]] = None

    def __init__(
        self,
        game_ids: Iterable[int],
        scores: np.ndarray,
        default_value: Optional[float] = None,
    ) -> None:
        self.default_value = (
            default_value if default_value is not None else scores.mean()
        )
        self.raw_scores = scores
        self.game_ids = tuple(game_ids)
        self.scores = defaultdict(
            self.default_factory,
            zip(self.game_ids, self.raw_scores),
        )

    @classmethod
    def train(cls, ratings: pd.DataFrame) -> "PopularGamesRecommender":
        """Train the recommender from ratings data."""
        raise NotImplementedError

    @classmethod
    def train_from_csv(cls, ratings_file: PATH) -> "PopularGamesRecommender":
        """Train the recommender from a ratings file in CSV format."""
        ratings = pd.read_csv(ratings_file)
        return cls.train(
            ratings[
                [
                    cls.id_field,
                    cls.user_id_field,
                    cls.rating_id_field,
                ]
            ]
        )

    @classmethod
    def train_from_json_lines(cls, ratings_file: PATH) -> "PopularGamesRecommender":
        """Train the recommender from a ratings file in JSON lines format."""
        ratings = pd.read_json(ratings_file, orient="records", lines=True)
        return cls.train(
            ratings[
                [
                    cls.id_field,
                    cls.user_id_field,
                    cls.rating_id_field,
                ]
            ]
        )

    @property
    def known_games(self) -> FrozenSet[int]:
        if self._known_games is not None:
            return self._known_games
        self._known_games = frozenset(self.game_ids)
        return self._known_games

    @property
    def rated_games(self) -> FrozenSet[int]:
        return self.known_games

    @property
    def num_games(self) -> int:
        return len(self.game_ids)

    @property
    def known_users(self) -> FrozenSet[str]:
        return frozenset()

    def default_factory(self) -> float:
        """Default value for unknown games."""
        return self.default_value

    def _recommendation_scores(
        self,
        users: int,
        games: Optional[List[int]] = None,
    ) -> np.ndarray:
        """Popularity scores."""
        scores = (
            np.array([self.scores[game_id] for game_id in games])
            if games
            else self.raw_scores
        )
        return np.tile(scores, [users, 1])

    def recommend(
        self,
        users: Iterable[str],
        **kwargs,
    ) -> pd.DataFrame:
        """Popular recommendations for certain users."""
        users = list(users)
        scores = self._recommendation_scores(users=len(users))
        return dataframe_from_scores(users, self.game_ids, scores)

    def recommend_as_numpy(
        self,
        users: Iterable[str],
        games: Iterable[int],
    ) -> np.ndarray:
        """Popular recommendations for certain users and games as a numpy array."""
        users = list(users)
        games = list(games)
        return self._recommendation_scores(users=len(users), games=games)

    def recommend_group(
        self,
        users: Iterable[str],
        **kwargs,
    ) -> pd.DataFrame:
        """Popular recommendations for a group of users."""
        scores = self._recommendation_scores(users=1)
        return dataframe_from_scores(["_all"], self.game_ids, scores)

    def recommend_group_as_numpy(
        self,
        users: Iterable[str],
        games: Iterable[int],
    ) -> np.ndarray:
        """Popular recommendations for a group of users and games as a numpy array."""
        games = list(games)
        return self._recommendation_scores(users=1, games=games)

    def recommend_similar(self, games: Iterable[int], **kwargs) -> pd.DataFrame:
        raise NotImplementedError

    def similar_games(self, games: Iterable[int], **kwargs) -> pd.DataFrame:
        raise NotImplementedError


class PopularMeanGamesRecommender(PopularGamesRecommender):
    """Recommend games by their mean rating score."""

    @classmethod
    def train(cls, ratings: pd.DataFrame) -> "PopularMeanGamesRecommender":
        data = ratings.groupby(cls.id_field, sort=False)[cls.rating_id_field].mean()
        return cls(
            game_ids=data.index,
            scores=data.to_numpy(),
            default_value=ratings[cls.rating_id_field].mean(),
        )


class PopularBayesianGamesRecommender(PopularGamesRecommender):
    """Recommend games by their Bayesian average rating score."""

    ratings_per_dummy: float = 10_000
    dummy_rating: Optional[float] = 5.5

    @classmethod
    def train(cls, ratings: pd.DataFrame) -> "PopularBayesianGamesRecommender":
        num_dummies = len(ratings) / cls.ratings_per_dummy
        dummy_rating = (
            ratings[cls.rating_id_field].mean()
            if cls.dummy_rating is None
            else cls.dummy_rating
        )

        stats = ratings.groupby(
            cls.id_field,
            sort=False,
        )[
            cls.rating_id_field
        ].agg(["size", "mean"])

        data = (stats["mean"] * stats["size"] + dummy_rating * num_dummies) / (
            stats["size"] + num_dummies
        )

        return cls(
            game_ids=data.index,
            scores=data.to_numpy(),
            default_value=dummy_rating,
        )


class PopularNumRatingsGamesRecommender(PopularGamesRecommender):
    """Recommend games by their number of ratings."""

    @classmethod
    def train(cls, ratings: pd.DataFrame) -> "PopularNumRatingsGamesRecommender":
        data = ratings.groupby(cls.id_field, sort=False).size()
        return cls(game_ids=data.index, scores=data.to_numpy())



"""Evaluate recommender models."""

import logging
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Iterable, Tuple, Union

import numpy as np
import polars as pl
from sklearn.metrics import ndcg_score

from board_game_recommender.base import BaseGamesRecommender

LOGGER = logging.getLogger(__name__)


@dataclass(frozen=True)
class RecommenderTestData:
    """Test data for recommender model evaluation."""

    user_ids: Tuple[str, ...]
    game_ids: np.ndarray
    ratings: np.ndarray


def load_test_data(
    path: Union[str, Path],
    ratings_per_user: int,
    user_id_key: str = "bgg_user_name",
    game_id_key: str = "bgg_id",
    ratings_key: str = "bgg_user_rating",
) -> RecommenderTestData:
    """Load RecommenderTestData from CSV."""

    path = Path(path).resolve()
    LOGGER.info("Loading test data from <%s>…", path)

    data = pl.read_csv(path)
    LOGGER.info("Read %d rows", len(data))

    if len(data) % ratings_per_user != 0:
        raise ValueError(
            f"The number of rows ({len(data)}) is not divisible by "
            + f"the number of ratings per user ({ratings_per_user})"
        )

    user_ids = tuple(data[user_id_key][::ratings_per_user])
    game_ids = data[game_id_key].view().reshape((-1, ratings_per_user))
    ratings = data[ratings_key].view().reshape((-1, ratings_per_user))

    return RecommenderTestData(user_ids=user_ids, game_ids=game_ids, ratings=ratings)


@dataclass(frozen=True)
class RecommenderMetrics:
    """Recommender model evaluation metrics."""

    ndcg: Dict[int, float]
    ndcg_exp: Dict[int, float]
    rmse: float
    effective_catalog_size: Dict[int, float]


def prediction_scores(
    recommender: BaseGamesRecommender,
    test_data: RecommenderTestData,
) -> np.ndarray:
    """Calculate the predicted scores from the recommender for the given test data."""
    return np.array(
        [
            recommender.recommend_as_numpy(users=(user,), games=games)[0, :]
            for user, games in zip(test_data.user_ids, test_data.game_ids)
        ]
    )


def effective_catalog_size(
    test_data: RecommenderTestData,
    y_pred: np.ndarray,
) -> np.ndarray:
    """Calculate the effective catalog size (ECS)."""

    assert len(test_data.user_ids) == len(y_pred)
    assert test_data.game_ids.shape == y_pred.shape

    pos_counts = (
        pl.LazyFrame(
            data={
                "user_id": np.repeat(test_data.user_ids, y_pred.shape[-1]),
                "game_id": test_data.game_ids.reshape(-1),
                "prediction": y_pred.reshape(-1),
            }
        )
        .with_columns(
            prediction_rank=pl.col("prediction")
            .rank(method="random", descending=True)
            .over("user_id")
        )
        .groupby("game_id", "prediction_rank")
        .count()
        .sort("game_id", "prediction_rank")
        .select(
            "game_id",
            "prediction_rank",
            pl.col("count").cumsum().over("game_id"),
        )
        .collect()
        .pivot(
            columns="game_id",
            index="prediction_rank",
            values="count",
            aggregate_function=None,
        )
        .lazy()
        .sort("prediction_rank")
        .fill_null(strategy="forward")
        .fill_null(0)
        .drop("prediction_rank")
        .collect()
        .to_numpy()
    )

    probs = pos_counts / pos_counts.sum(axis=1).reshape((-1, 1))
    ranks = np.argsort(-1 * pos_counts).argsort() + 1

    return 2 * np.sum(probs * ranks, axis=1) + 1


def calculate_metrics(
    recommender: BaseGamesRecommender,
    test_data: RecommenderTestData,
    *,
    k_values: Union[None, int, Iterable[int]],
) -> RecommenderMetrics:
    """Calculate RecommenderMetrics for given recommender model and RecommenderTestData."""

    y_true = test_data.ratings
    y_pred = prediction_scores(recommender, test_data)

    if y_true.shape != y_pred.shape:
        raise ValueError(
            f"Shape of ratings ({y_true.shape}) does not match "
            + f"shape of predictions ({y_pred.shape})"
        )

    rmse = float(np.sqrt(np.square(y_true - y_pred).mean()))

    if k_values is None:
        k_values = frozenset()
    elif isinstance(k_values, int):
        k_values = frozenset({k_values})
    else:
        k_values = frozenset(k_values)

    k_values = sorted(k_values | {y_true.shape[-1]})

    ecs_all = effective_catalog_size(test_data, y_pred)
    ecs = {k: ecs_all[k - 1] for k in k_values}

    ndcg = {}

    for k in k_values:
        ndcg[k] = ndcg_score(
            y_true=y_true,
            y_score=y_pred,
            k=k,
        )

    y_true = np.exp2(y_true) - 1
    ndcg_exp = {}

    for k in k_values:
        ndcg_exp[k] = ndcg_score(
            y_true=y_true,
            y_score=y_pred,
            k=k,
        )

    return RecommenderMetrics(
        ndcg=ndcg,
        ndcg_exp=ndcg_exp,
        rmse=rmse,
        effective_catalog_size=ecs,
    )
